/* main.c */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "X11/Xlib.h"
#include "X11/Xutil.h"
#include <stdio.h>
#include <fcntl.h>
#include <sys/lock.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

int brgc;
char *brgv[100], cc[1000];
int Argc;
char **Argv;


static char
  ProgramName[128];

int *shmw;

/* call to link to shared memory */
char *mcshmat();
char *malloc();
Display *dpy;

static int novideo = -1234567;

/******************************************************************************/
main(argc,argv)
int argc;
char *argv[];
{
   static int zero=0, one=1;
   int i, depth, nitems;
   char *ap, *cp;

   Argc = argc;
   Argv = argv;

   brgc = argc;
   cp = cc;
   for (i=0; i<argc; i++) {
      brgv[i] = cp;
      ap = argv[i];
      while (*ap != '\0') {
         if (*ap >= 'a' && *ap <= 'z')
            *cp++ = (*ap++) + 'A' - 'a';
         else
            *cp++ = *ap++;
      }
      *cp++ = '\0';
   }

   /* Set program name */
   strcpy(ProgramName, brgv[0]);

   shmw = (int*)mcshmat(0);

#ifdef ibm
   i=main0();
#else
   i=main0_();
#endif
   printf(ProgramName+17);
   printf(" done\n");
   return(0);
}

/******************************************************************************/
#define SUBSTR(sub, str) (memcmp(sub, str, strlen(sub)) ? 0 : 1)

#ifdef ibm
  nkwp(ckw,lenkw)
#else
  nkwp_(ckw,lenkw)
#endif
int lenkw;
char *ckw;
{
  char ck[20],*cmd;
  int i,j,n;
  strncpy(ck,ckw,lenkw);
  ck[lenkw]=0;

  if (strcmp(ck, " ") == 0) {
    j = 1;
    n = 0;
    cmd = brgv[j];
    while (j < brgc) {
      if (cmd[0] == '-' && (cmd[1] < '0' || cmd[1] > '9'))
        break;
      cmd = brgv[++j];
      n++;
    }
  }
  else {
    n = 0;
    for (i = 1; i < brgc; i++) {
      cmd = brgv[i];
      if (cmd[0] == '-' && (cmd[1] < '0' || cmd[1] > '9') &&
                           SUBSTR(cmd + 1, ck)) {
        j = i + 1;
        n = 0;
        cmd = brgv[j];
        while (j < brgc) {
          if (cmd[0] == '-' && (cmd[1] < '0' || cmd[1] > '9'))
            break;
          cmd = brgv[++j];
          n++;
        }
        if (n == 0) n = 1;
        break;
      }
    }
  }
  return(n);
}

/* cres = result string, len = length limit for cres
   ckw = keyword string, of length lenkw
   ip = parameter # after keyword
   cin = default value of length lenc
   NOTE strings are FORTRAN strings, so no
   null termination and blank padded to length
*/
#ifdef ibm
  ckwp(cres,len,ckw,ip,cin,lenkw,lenc)
#else
  ckwp_(cres,len,ckw,ip,cin,lenkw,lenc)
#endif
int *ip,len,lenkw,lenc;
char *cres,*ckw,*cin;
{
  char *d,c[20],ck[20],*cmd;
  int i,j,k,n;
/* printf("cpp: %d %d %s %d\n",len,ip[0],cin,lenc); */
  strncpy(c,cin,lenc);
  c[lenc]=0;
  strncpy(ck,ckw,lenkw);
  ck[lenkw]=0;

  d=c;
  if (strcmp(ck, " ") == 0) {
    j = 1;
    n = 1;
    cmd = brgv[j];
    while (j < brgc) {
      if (cmd[0] == '-' && (cmd[1] < '0' || cmd[1] > '9'))
        break;
      if (n == ip[0]) {
        d = brgv[j];
        break;
      }
      cmd = brgv[++j];
      n++;
    }
  }
  else {
    for (i = 1; i < brgc; i++) {
      cmd = brgv[i];
      if (cmd[0] == '-' && (cmd[1] < '0' || cmd[1] > '9') &&
                           SUBSTR(cmd + 1, ck)) {
        j = i + 1;
        n = 1;
        cmd = brgv[j];
        while (j < brgc) {
          if (cmd[0] == '-' && (cmd[1] < '0' || cmd[1] > '9'))
            break;
          if (n == ip[0]) {
            d = brgv[j];
            break;
          }
          cmd = brgv[++j];
          n++;
        }
        break;
      }
    }
  }
  k=strlen(d);
  if (len<k)
    k=len;
  for (j=0; j<k; j++)
    cres[j]=d[j];
  for (j=k; j<len; j++)
    cres[j]=' ';
  return(0);
}

/******************************************************************************/
#define NFILES 8
int ld=0,nlw=0;
char f[NFILES][256];
int fd[NFILES];

#ifdef ibm
  lwi(cf,iw,nw,ibuf,length)
#else
  lwi_(cf,iw,nw,ibuf,length)
#endif
int *iw,*nw,*ibuf,length;
char *cf;
{
   int nbytes,ifile,iword,nwords,len,kword;
   long offset;
   char cfile[256];
   len = length;
   strncpy(cfile,cf,len);
   cfile[len]=0;
   len = len-1;
   while (cfile[len] == ' ' && len > 0) {
      cfile[len] = 0;
      len = len-1;
   }
   iword=iw[0];
   nwords=nw[0];
   if ( (ifile=lwopen(cfile, 1)) <0) {
      for ( kword = 0; kword < nwords; kword++)
         ibuf[kword] = 0x80808080;
      return(-1);
   }
   offset=4*iword;
  
   if (lseek(ifile,offset,0)<0) {
      for ( kword = 0; kword < nwords; kword++)
         ibuf[kword] = 0x80808080;
      return(-2);
   }

   nbytes=4*nwords;
   if (read(ifile,ibuf,nbytes)<=0) {
      for ( kword = 0; kword < nwords; kword++)
         ibuf[kword] = 0x80808080;
        return(-3);
   }
   return(0);
}

#ifdef ibm
  lwo(cf,iw,nw,ibuf,length)
#else
  lwo_(cf,iw,nw,ibuf,length)
#endif
int *iw,*nw,length;
char *cf,*ibuf;
{
   int nbytes,ifile,iword,nwords,len,kword;
   int nblocks, iblk;
   long offset,filelen, gapword;
   char cfile[256];
   long int nullbuf[512];
   static int nbuf = 512;
   len = length;
   strncpy(cfile,cf,len);
   cfile[len]=0;
   len = len-1;
   while (cfile[len] == ' ' && len > 0) {
      cfile[len] = 0;
      len = len-1;
   }
   iword=iw[0];
   nwords=nw[0];
   if ( (ifile=lwopen(cfile, 0)) <0)
      return(-1);
   offset=4*iword;
   filelen = lseek(ifile, (long) 0, 2);
/* check to see if we write past eof */
   if (offset > filelen) {
      for (kword = 0; kword < nbuf; kword++) 
         nullbuf[kword] = 0x80808080;
      gapword = (offset - filelen)/4;
      nblocks = gapword/nbuf;
      if( nblocks != 0) {
          nbytes = nbuf * 4;
          for ( iblk = 0; iblk < nblocks; iblk++)
              write(ifile, nullbuf, nbytes);
      }
      nbytes = 4 * (gapword % nbuf);
      if ( nbytes != 0 )
          write(ifile, nullbuf, nbytes);
   }
   if (lseek(ifile,offset,0)<0)
      return(-2);
   nbytes=4*nwords;
   if (write(ifile,ibuf,nbytes)<0)
      return(-3);
   return(0);
}

#ifdef ibm
  lbi(cf,ib,nb,ibuf,length)
#else
  lbi_(cf,ib,nb,ibuf,length)
#endif
int *ib,*nb,length;
char *cf,*ibuf;
{
   int nbytes,ifile,len;
   long offset;
   char cfile[256];
   len = length;
   strncpy(cfile,cf,len);
   cfile[len]=0;
   len = len-1;
   while (cfile[len] == ' ' && len > 0) {
      cfile[len] = 0;
      len = len-1;
   }
   if ( (ifile=lwopen(cfile, 1)) <0) {
      ibuf[0] = 0X80;
      return(-1);
   }
   offset=ib[0];
  
   if (lseek(ifile,offset,0)<0)
      return(-2);
   nbytes=nb[0];
   if (read(ifile,ibuf,nbytes)<=0)
        return(-3);
   return(0);
}

#ifdef ibm
  lbo(cf,ib,nb,ibuf,length)
#else
  lbo_(cf,ib,nb,ibuf,length)
#endif
int *ib,*nb,length;
char *cf,*ibuf;
{
   int nbytes,ifile,len;
   long offset;
   char cfile[256];
   len = length;
   strncpy(cfile,cf,len);
   cfile[len]=0;
   len = len-1;
   while (cfile[len] == ' ' && len > 0) {
      cfile[len] = 0;
      len = len-1;
   }
   if ( (ifile=lwopen(cfile, 0)) <0)
      return(-1);
   offset=ib[0];
   if (lseek(ifile,offset,0)<0)
      return(-2);
   nbytes=nb[0]; 
   if (write(ifile,ibuf,nbytes)<0)
      return(-3);
   return(0);
}

lwopen(cfile, rdflag)
char *cfile;
int rdflag;
{
   int id,j,len;
   len=strlen(cfile);
/* printf("\nfile: %s length: %d\n",cfile,len); */
   for (id=0; id<nlw; id++) {
      for (j=0; j<len; j++) {
         if (cfile[j]!=f[id][j])
            goto notf;
      }
      return(fd[id]);
notf: ;
   }
   if(ld < nlw) close(fd[ld]);
   for (j=0; j<=len; j++)
      f[ld][j]=cfile[j];
   umask(0);
   if (rdflag) fd[ld]=open(cfile,O_RDWR,0666);
   else fd[ld]=open(cfile,O_RDWR|O_CREAT,0666);
   if (fd[ld]<0) {
      if (!rdflag) printf("\ncannot open file: %s\n",cfile);
      return(-1);
   }
   id=ld;
   ++ld;
   if (ld>nlw)
      nlw=ld;
   if (ld>NFILES-1)
      ld=0;
   return(fd[id]);
}

/******************************************************************************/
#ifdef ibm
  cdest(c,i,len)
#else
  cdest_(c,i,len)
#endif
int *i,len;
char *c;
{
   int j;
   c[len]=0;
   j=i[0];
   printf(c);
   if (j!=0)
      printf("%8d\n",j);
   else
      printf("\n");
   return(0);
}

#ifdef ibm
  movb(n,c,d,k)
#else
  movb_(n,c,d,k)
#endif
int *n,*k;
char *c,*d;
{
   int i,j;
   for (i=0; i<n[0]; i++)
      d[k[0]+i]=c[i];
   return(0);
}

/******************************************************************************/

#ifdef ibm
  flock(i)
#else
  flock_(i)
#endif
int *i;
{
/* for this to work, must do the following to the command
   calling flock_() :
    chown root command
    chgrp root command
    chmod u+s command
*/
   if (setuid(0) != 0)
      printf("bad setuid %d\n", errno);
   if (i[0] == 0)
      if (plock(PROCLOCK) != 0)
         printf("bad plock %d\n", errno);
   else
      if (plock(UNLOCK) != 0)
         printf("bad plock %d\n", errno);
   return(0);
}


#define SHMKEY 123

char *shmad;
int shmid;

char* mcshmat(size)
int size;
{
   char *shmat();
   int shmget(), intad;
   shmid = shmget(SHMKEY, 4*size, (IPC_CREAT | 0666));
   if(shmid == -1 && size != 0) {
      shmid = shmget(SHMKEY, 0, (IPC_CREAT | 0666));
      shmad = shmat(shmid, 0, 0);
      mcshmrm();
      shmid = shmget(SHMKEY, 4*size, (IPC_CREAT | 0666));
      if(shmid == -1) {
         printf("unable to shmget %d %d\n", errno, 4*size);
         exit(0);
      }
   }
   if(shmid == -1) {
      shmad = (char*)&novideo;
      return(shmad);
   }
   shmad = shmat(shmid, 0, 0);
   intad = (int)shmad;
   if(intad == -1) {
      printf("unable to shmat %d %d\n", errno, 4*size);
      exit(0);
   }
   return(shmad);
}

mcshmrm()
{
   struct shmid_ds *buf;
   shmctl(shmid, IPC_RMID, buf);
}

#ifdef ibm
  mckill(mcset)
#else
  mckill_(mcset)
#endif
int *mcset;
{
    int mcwas;
    mcwas = shmw[0];
    shmw[0] = *mcset;
    *mcset = mcwas;
    shmw[12] = 1;
    if(*mcset < 0) {
       sleep(5);
       mcshmrm();
    }
}

/******************************************************************************/
#ifdef ibm
  luc(index)
#else
  luc_(index)
#endif
   int *index;
{
   return(0);
}

#ifdef ibm
  puc(i, index)
#else
  puc_(i, index)
#endif
   int *i, *index;
{
   return(0);
}



#ifdef ibm
/* IDATE not supported on IBM? */
idate( imon, iday, iyear )
int *imon, *iday, *iyear;
{
   *imon = 1;
   *iday = 1;
   *iyear = 92;
   return 0;
}
#endif


